"""Download a list of files to given location."""
import asyncio
import logging
import pathlib
from typing import List, Tuple, Union

import aiofiles
import aiohttp
import async_timeout

from mariner import utils

Url = str
Path = Union[str, pathlib.Path]
File = Tuple[Url, str]
Loop = asyncio.selector_events.BaseSelectorEventLoop
Session = aiohttp.ClientSession


class Downloader:
    """Handle file downloads asynchronously."""

    log = logging.getLogger(__name__)

    def __init__(self, download_path: Path = '~/Downloads') -> None:
        self.download_path = utils.check_path(download_path)

    async def download_coroutine(self, session: Session, url: Url, name: str) -> Path:
        """Download a single file and asynchronously save it to disk.

        Args:
            url: URL of file.
            name: Name of file.

        Returns:
            Path of downloaded file.
        """
        filename = self.download_path / name
        self.log.debug('filename=%s', filename)
        with async_timeout.timeout(10):
            async with session.get(url) as response:
                async with aiofiles.open(filename, 'wb') as file_:
                    while True:
                        chunk = await response.content.read(1024)
                        if not chunk:
                            self.log.debug('saved=%s', filename)
                            break
                        await file_.write(chunk)
        return filename

    async def download_filelist(self, loop: Loop, download_list: List[File]) -> List[Path]:
        """Download all files in the list asychronously.

        Args:
            loop: AsyncIO loop.
            filelist: List of files to download.

        Returns:
            List of downloaded files.
        """
        filelist = []
        async with aiohttp.ClientSession(loop=loop) as session:
            for url, name in download_list:
                self.log.debug('Downloading url=%s name=%s', url, name)
                filelist.append(await self.download_coroutine(session, url, name))
        return filelist

    def download(self, download_list: List[File]) -> List[Path]:
        """Run a loop to download all files in the list.

        Args:
            download_list: List of files to download.

        Return:
            List of downloaded files.
        """
        loop = asyncio.get_event_loop()
        filelist = loop.run_until_complete(self.download_filelist(loop, download_list))
        return filelist



# -*- coding: future_fstrings -*-
"""Mariner download command."""
import logging

from cliff import lister

from mariner import downloader, utils


class Download(lister.Lister):
    """Download torrent with given ID."""

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        """Add arguments to argument parser.

        Args:
            prog_name: Application name.

        Returns:
            Instance of argument parser.
        """
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'ID', nargs='+', help="ID of the torrent to download", type=int)
        return parser

    def take_action(self, parsed_args):
        """Download chosen torrent.

        Args:
            parsed_args: List of parsed arguments.

        Returns:
            List of downloaded torrents and the location where they were saved.
        """
        torrents = []
        for tid in parsed_args.ID:
            torrent_ = self.app.engine.result(tid)
            self.log.debug('tid=%s torrent=%s', tid, torrent_)
            if torrent_.torrent:
                torrents.append(torrent_)
                self.log.debug('Torrent appended.')
                self.log.info(f'Downloading torrent ID {utils.cyan(tid)}.')
            else:
                self.log.warning(utils.yellow(
                    f'No torrent for {torrent_.name}. Use magnet link instead.'))

        download_list = ((t.torrent, t.filename) for t in torrents)
        path = self.app.config['download_path']
        self.log.debug('download_list=%s download_path=%s', download_list, path)
        torrent_downloader = downloader.Downloader(download_path=path)
        filelist = torrent_downloader.download(download_list)

        headers = ('Name', 'Saved to')
        colored_headers = [utils.magenta(h) for h in headers]
        columns = zip((t.colored().name for t in torrents), filelist)
        return (colored_headers, columns)



# -*- coding: future_fstrings -*-
"""Mariner search command."""
import logging

from cliff import lister

from mariner import torrent, utils


class Search(lister.Lister):
    """Search for torrents."""

    log = logging.getLogger(__name__)

    @staticmethod
    def _availability(torrent_: torrent.Torrent) -> str:
        """Show whether it is available as torrent, magnet link or both."""
        availability = []
        if torrent_.magnet:
            availability.append('Magnet link')
        if torrent_.torrent:
            availability.append('Torrent')
        return ', '.join(availability)

    def get_parser(self, prog_name):
        """Add arguments to argument parser.

        Args:
            prog_name: Application name.

        Returns:
            Instance of argument parser.
        """
        parser = super().get_parser(prog_name)
        parser.add_argument('title', help='Title to search for')
        parser.add_argument('--all', '-a', action='store_true',
                            help='Search all available trackers')
        parser.add_argument('--limit', '-l', nargs='?',
                            default=self.app.config['results'],
                            help='Limit the number of results shown. Default is 50.',
                            type=int)
        parser.add_argument('--newest', '-n', action='store_true',
                            help='Sort results by newest')
        plugins = self.app.engine.plugins.keys()
        parser.add_argument('--trackers', '-t', action='append', choices=plugins,
                            help='Trackers that should be searched', default=[])
        return parser

    def take_action(self, parsed_args):
        """Search for a torrent.

        Args:
            parsed_args: List of parsed arguments.

        Returns:
            List of torrents found.
        """
        title = parsed_args.title
        limit = parsed_args.limit
        newest = parsed_args.newest

        if parsed_args.all:
            # Use all trackers
            trackers = self.app.engine.plugins.keys()
        else:
            # If default tracker is used as default argument, the user provided ones
            # are appended to it, instead of replacing it.
            if not parsed_args.trackers:
                parsed_args.trackers.append(self.app.config['default_tracker'])
            trackers = [t.lower() for t in set(parsed_args.trackers)]

        self.log.info(f'Searching for {utils.cyan(title)}.')
        self.log.debug('title=%s limit=%s trackers=%s', title, limit, trackers)
        results = self.app.engine.search(
            title, trackers, limit, sort_by_newest=newest)

        headers = ('ID', 'Name', 'Tracker', 'Seeds',
                   'Size', 'Uploaded', 'Available as')
        # Headers cannot be a generator, as it messes up alignment
        colored_headers = [utils.magenta(h) for h in headers]
        colored_results = ((tid, t.colored()) for tid, t in results)
        columns = ((tid,
                    t.name,
                    t.tracker,
                    t.seeds,
                    t.size,
                    t.date,
                    self._availability(t))
                   for tid, t in colored_results)
        return (colored_headers, columns)



"""Module for searching torrents on KickAssTorrents."""
from typing import Iterator

import bs4

from mariner import searchengine, torrent


class KickAssTorrents(searchengine.TrackerPlugin):
    """Represents KickAssTorrents search engine."""

    search_url = 'https://katcr.co/katsearch/page/1/'
    aliases = ['kat']

    def _parse(self, raw: str) -> Iterator[torrent.Torrent]:  # pylint: disable=too-many-locals
        """Parse result page.

        Args:
          raw: Raw HTML results page to parse.

        Returns:
            List of torrent names with magnet links.
        """
        soup = bs4.BeautifulSoup(raw, 'lxml')
        contents = soup.find('table', class_='torrents_table')
        try:
            for line in contents.find_all('tr')[1:]:
                torrent_ = line.find(
                    'div', class_='torrents_table__torrent_name')
                name = str(torrent_.find(
                    'a', class_='torrents_table__torrent_title').string)
                magnet = torrent_.find(
                    'a', {'title': 'Torrent magnet link'})['href']
                tracker = self.__class__.__name__
                size = line.find('td', {'data-title': 'Size'}).get_text()
                date = line.find('td', {'data-title': 'Age'}).get_text()
                raw_seeds = line.find('td', {'data-title': 'Seed'}).string
                seeds = self._parse_number(raw_seeds)
                raw_leeches = line.find('td', {'data-title': 'Leech'}).string
                leeches = self._parse_number(raw_leeches)
                yield torrent.Torrent(name, tracker, magnet=magnet, size=size,
                                      date=date, seeds=seeds, leeches=leeches)
        except AttributeError:
            self.log.debug("No results found")
            yield from ()



# -*- coding: future_fstrings -*-
"""Handle searching for torrents on torrent trackers."""
import abc
import asyncio
import importlib
import itertools
import logging
import pathlib
from typing import Iterator, List, Optional, Tuple, Union

import aiohttp
import async_timeout
import cachalot

from mariner import torrent, exceptions

Name = str
Page = str
Path = Union[str, pathlib.Path]
Url = str


class SearchEngine:
    """Search on for a torrent using Tracker plugins."""

    log = logging.getLogger(__name__)
    plugin_directory = pathlib.Path(__file__).parent / 'plugins'

    def __init__(self) -> None:
        self.plugins = {}
        self.results = cachalot.Cache(
            path='~/.local/share/mariner/results.json', size=1000)
        self.initialize_plugins()

    @staticmethod
    def _flatten(nested_list: List[List]) -> List:
        """Flatten a list."""
        return list(itertools.chain(*nested_list))

    def find_plugins(self) -> None:
        """Find and import tracker plugins."""
        for module in self.plugin_directory.glob('*.py'):  # pylint: disable=no-member
            self.log.debug('Loading module=%s', module)
            name = module.stem
            spec = importlib.util.spec_from_file_location(name, module)
            loaded_module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(loaded_module)

    def initialize_plugins(self) -> None:
        """Find engines and register them."""
        self.log.debug('Initializing plugins')
        self.find_plugins()
        for plugin in TrackerPlugin.__subclasses__():
            self.log.debug('Adding plugin=%s', plugin)
            name = plugin.__name__.lower()
            self.plugins[name] = plugin
            for alias in plugin.aliases:
                self.plugins[alias] = plugin

    def result(self, tid: str) -> torrent.Torrent:
        """Get torrent of given id.

        Args:
            tid: ID of the torrent to get.

        Returns:
            Torrent with given ID.
        """
        self.log.debug('Fetching torrent with tid=%s', tid)
        torrent_ = self.results.get(tid)
        if torrent_:
            return torrent_
        raise exceptions.NoResultException(f"No torrent with ID {tid}")

    @cachalot.Cache(path='.cache/mariner/cache.json', size=100)
    def _cached_search(self,
                       title: str,
                       trackers: List[str],
                       ) -> List[torrent.Torrent]:  # pylint: disable=bad-continuation
        """Search for torrents on given site and cache to results. This
        method is an implementation detail. As coroutines are not easily
        serializable, we cannot simply cache TrackerPlugun.results() method.

        Args:
            title: String to search for.

        Returns:
            List of Torrents returned by the search.
        """
        self.log.debug('Fetching search results')

        # Get unique trackers
        try:
            plugins = set(self.plugins[t] for t in trackers)
        except KeyError:
            raise exceptions.ConfigurationError("Illegal value for default_tracker")
        else:
            tasks = asyncio.gather(*(p().results(title.lower()) for p in plugins))
            loop = asyncio.get_event_loop()
            torrents = loop.run_until_complete(tasks)
            torrents = self._flatten(torrents)
        return torrents

    def search(self,
               title: str,
               trackers: List[str],
               limit: Optional[int] = 10,
               sort_by_newest: bool = False,
               ) -> List[Tuple[int, torrent.Torrent]]:  # pylint: disable=bad-continuation
        """Search for torrents on given site.

        Args:
            title: String to search for.
            limit: Defaults to 10. Limits the number of results, that should be shown.

        Returns:
            List of Torrents returned by the search, up to the limit.
        """
        if not title:
            raise exceptions.InputError('No string to search for.')
        if not trackers:
            raise exceptions.InputError('No torrent trackers to search on')
        if limit <= 0:
            raise exceptions.InputError('Limit has to be higher than zero.')

        torrents = self._cached_search(title, trackers)

        # Sort the results
        if sort_by_newest:
            sorted_torrents = list(reversed(sorted(torrents, key=lambda x: x.date)))
        else:
            sorted_torrents = list(reversed(sorted(torrents)))

        # Show only results up to to the limit
        results = [(i, t) for i, t in enumerate(sorted_torrents[:limit])]
        if results:
            self.save_results(results)
            return results
        raise exceptions.NoResultException(f"No results for {title}")

    def save_results(self, torrents: List[Tuple[int, torrent.Torrent]]) -> None:
        """Save results in a database.

        Args:
            torrents: List of ID, Torrent tuples.
        """
        self.results.clear()
        for tid, torrent_ in torrents:
            self.results.insert(tid, torrent_)


class TrackerMeta(abc.ABCMeta, type):
    """Metaclass to check, that Tracket plugins override search_url."""
    def __new__(mcs, name, bases, namespace, **kwargs):
        if bases != (abc.ABC,):
            if not namespace.get('search_url'):
                raise exceptions.InputError('You must define search_url')
        return type.__new__(mcs, name, bases, namespace)


class TrackerPlugin(abc.ABC, metaclass=TrackerMeta):
    """Represent a search engine."""
    log = logging.getLogger(__name__)
    user_agent = {'user-agent': 'Mariner Torrent Downloader'}
    search_url = ''  # To be overwritten by subclasses
    aliases = []  # Aliases for the tracker name

    async def get(self, url: Url) -> Page:
        """Asynchronous https request.

        Args:
            url: Url of the webpage to fetch.

        Returns:
            Fetched webpage."""
        self.log.debug('Fetching url=%s', url)
        with async_timeout.timeout(10):
            async with aiohttp.ClientSession() as session:
                async with session.get(url, headers=self.user_agent) as response:
                    page = await response.text()
        return page

    async def results(self, title: str) -> Iterator[torrent.Torrent]:
        """Get a list of torrent name with URLs and magnet links.

        Args:
            title: String to search for.
        """
        try:
            search_url = self.search_url + title
            page = await self.get(search_url)
        except (OSError, asyncio.TimeoutError):
            self.log.error('Cannot reach server at %s', search_url)
        return self._parse(page)

    @abc.abstractmethod
    def _parse(self, raw: str) -> Iterator[torrent.Torrent]:
        """Parse result page.

        Args:
            raw: Raw HTML page.

        Returns:
            List of torrents names with URLs and magnet links.

        """
        raise NotImplementedError

    @staticmethod
    def _parse_number(number: str) -> int:
        """Parse a number string from HTML page and return an interger.

        Args:
            number: Number string to parse.

        Return:
            Parsed number.
        """
        squashed = number.replace(' ', '')
        return int(squashed.replace(',', ''))



"""Module for searching torrents on PirateBay."""
from typing import Iterator

import bs4

from mariner import searchengine, torrent


class PirateBay(searchengine.TrackerPlugin):
    """Represents PirateBay search engine."""

    search_url = 'https://thepiratebay.org/search/'
    aliases = ['tpb', 'pb']

    def _parse(self, raw: str) -> Iterator[torrent.Torrent]:  # pylint: disable=too-many-locals
        """Parse result page.

        Args:
          raw: Raw HTML results page to parse.

        Returns:
            List of torrent names with magnet links.
        """
        soup = bs4.BeautifulSoup(raw, 'lxml')
        content = soup.find('table', id='searchResult')
        try:
            torrents = content.find_all('tr')
        except AttributeError:
            # No search result
            yield from ()
        else:
            for torrent_ in torrents[1:]:
                raw_name = torrent_.find('a', class_='detLink')
                name = PirateBay._parse_name(raw_name)
                links = torrent_.find_all('a')
                magnet = links[3].get('href')
                tracker = self.__class__.__name__

                numbers = torrent_.find_all('td', align='right')
                raw_seeds = numbers[0].string
                seeds = self._parse_number(raw_seeds)
                raw_leeches = numbers[1].string
                leeches = self._parse_number(raw_leeches)

                description = torrent_.find(
                    'font', class_="detDesc").get_text()
                fields = description.split(',')
                date = ' '.join(fields[0].split()[1:])
                size = ' '.join(fields[1].split()[-2:])

                yield torrent.Torrent(name, tracker, magnet=magnet, seeds=seeds,
                                      leeches=leeches, date=date, size=size)

    @staticmethod
    def _parse_name(raw: bs4.NavigableString) -> str:
        """Parse torrent name.

        Args:
          raw: Raw name to parse.

        Returns:
            Parsed name.
        """
        name = raw.string
        if name is None:
            name = raw['title'].split(" ")[2:]
            name = ' '.join(str(x) for x in name)
        return name.strip()
